/***************************************************************************
                          tnlTraverser_Grid3D_impl.h  -  description
                             -------------------
    begin                : Jul 29, 2014
    copyright            : (C) 2014 by Tomas Oberhuber
    email                : tomas.oberhuber@fjfi.cvut.cz
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef TNLTRAVERSER_GRID3D_IMPL_H_
#define TNLTRAVERSER_GRID3D_IMPL_H_

template< typename Real,
          typename Index >
   template< typename UserData,
             typename EntitiesProcessor >
void
tnlTraverser< tnlGrid< 3, Real, tnlHost, Index >, 3 >::
processBoundaryEntities( const GridType& grid,
                         UserData& userData ) const
{
   /****
    * Boundary conditions
    */
   const int CellDimensions = GridType::Dimensions;
   typename GridType::template GridEntity< CellDimensions > entity;
   
   CoordinatesType& coordinates = entity.getCoordinates();
   const IndexType& xSize = grid.getDimensions().x();
   const IndexType& ySize = grid.getDimensions().y();
   const IndexType& zSize = grid.getDimensions().z();

   for( coordinates.y() = 0; coordinates.y() < ySize; coordinates.y() ++ )
      for( coordinates.x() = 0; coordinates.x() < xSize; coordinates.x() ++ )
      {
         coordinates.z() = 0;
         EntitiesProcessor::processEntity( grid, userData, grid.getEntityIndex( entity ), entity );
         coordinates.z() = zSize - 1;
         EntitiesProcessor::processEntity( grid, userData, grid.getEntityIndex( entity ), entity );
      }

   for( coordinates.z() = 0; coordinates.z() < zSize; coordinates.z() ++ )
      for( coordinates.x() = 0; coordinates.x() < xSize; coordinates.x() ++ )
      {
         coordinates.y() = 0;
         EntitiesProcessor::processEntity( grid, userData, grid.getEntityIndex( entity ), entity );
         coordinates.y() = ySize - 1;
         EntitiesProcessor::processEntity( grid, userData, grid.getEntityIndex( entity ), entity );
      }

   for( coordinates.z() = 0; coordinates.z() < zSize; coordinates.z() ++ )
      for( coordinates.y() = 0; coordinates.y() < ySize; coordinates.y() ++ )
      {
         coordinates.x() = 0;
         EntitiesProcessor::processEntity( grid, userData, grid.getEntityIndex( entity ), entity );
         coordinates.x() = xSize - 1;
         EntitiesProcessor::processEntity( grid, userData, grid.getEntityIndex( entity ), entity );
      }
}
template< typename Real,
          typename Index >
   template< typename UserData,
             typename EntitiesProcessor >
void
tnlTraverser< tnlGrid< 3, Real, tnlHost, Index >, 3 >::
processInteriorEntities( const GridType& grid,
                         UserData& userData ) const
{
   /****
    * Interior cells
    */
   typedef typename GridType::Cell CellTopology;
   CoordinatesType coordinates;
   const IndexType& xSize = grid.getDimensions().x();
   const IndexType& ySize = grid.getDimensions().y();
   const IndexType& zSize = grid.getDimensions().z();

#ifdef HAVE_OPENMP
//#pragma omp parallel for
#endif
   for( coordinates.z() = 1; coordinates.z() < zSize - 1; coordinates.z() ++ )
      for( coordinates.y() = 1; coordinates.y() < ySize - 1; coordinates.y() ++ )
         for( coordinates.x() = 1; coordinates.x() < xSize - 1; coordinates.x() ++ )
         {
            const IndexType index = grid.getCellIndex( coordinates );
            EntitiesProcessor::template processEntity< CellTopology >( grid, userData, index, coordinates );
         }
}

template< typename Real,
          typename Index >
   template< typename UserData,
             typename EntitiesProcessor >
void
tnlTraverser< tnlGrid< 3, Real, tnlHost, Index >, 2 >::
processBoundaryEntities( const GridType& grid,
                         UserData& userData ) const
{
   /****
    * Traversing boundary faces
    */
   CoordinatesType coordinates;
   const IndexType& xSize = grid.getDimensions().x();
   const IndexType& ySize = grid.getDimensions().y();
   const IndexType& zSize = grid.getDimensions().z();

   for( coordinates.y() = 0; coordinates.y() < ySize; coordinates.y() ++ )
      for( coordinates.x() = 0; coordinates.x() < xSize; coordinates.x() ++ )      
      {
         typedef typename GridType::template Face< 0, 0, 1 > FaceTopology;
         coordinates.z() = 0;
         EntitiesProcessor::template processEntity< FaceTopology >( grid, userData, grid.template getFaceIndex< 0, 0, 1 >( coordinates ), coordinates );
         //cout << "Boundary face coordinates = " << coordinates << " index = " << grid.template getFaceIndex< 0, 1 >( coordinates ) << endl;
         coordinates.z() = zSize;
         EntitiesProcessor::template processEntity< FaceTopology >( grid, userData, grid.template getFaceIndex< 0, 0, 1 >( coordinates ), coordinates );
         //cout << "Boundary face coordinates = " << coordinates << " index = " << grid.template getFaceIndex< 0, 1 >( coordinates ) << endl;
      }
     
   for( coordinates.z() = 0; coordinates.z() < zSize; coordinates.z() ++ )
      for( coordinates.x() = 0; coordinates.x() < xSize; coordinates.x() ++ )
      {
         typedef typename GridType::template Face< 0, 1, 0 > FaceTopology;
         coordinates.y() = 0;
         EntitiesProcessor::template processEntity< FaceTopology >( grid, userData, grid.template getFaceIndex< 0, 1, 0 >( coordinates ), coordinates );
         //cout << "Boundary face coordinates = " << coordinates << " index = " << grid.template getFaceIndex< 0, 1 >( coordinates ) << endl;
         coordinates.y() = ySize;
         EntitiesProcessor::template processEntity< FaceTopology >( grid, userData, grid.template getFaceIndex< 0, 1, 0 >( coordinates ), coordinates );
         //cout << "Boundary face coordinates = " << coordinates << " index = " << grid.template getFaceIndex< 0, 1 >( coordinates ) << endl;
      }
  
   for( coordinates.z() = 0; coordinates.z() < zSize; coordinates.z() ++ )
      for( coordinates.y() = 0; coordinates.y() < ySize; coordinates.y() ++ )
      {
         typedef typename GridType::template Face< 1, 0, 0 > FaceTopology;
         coordinates.x() = 0;
         EntitiesProcessor::template processEntity< FaceTopology >( grid, userData, grid.template getFaceIndex< 1, 0, 0 >( coordinates ), coordinates );
         //cout << "Boundary face coordinates = " << coordinates << " index = " << grid.template getFaceIndex< 1, 0 >( coordinates ) << endl;
         coordinates.x() = xSize;
         EntitiesProcessor::template processEntity< FaceTopology >( grid, userData, grid.template getFaceIndex< 1, 0, 0 >( coordinates ), coordinates );
         //cout << "Boundary face coordinates = " << coordinates << " index = " << grid.template getFaceIndex< 1, 0 >( coordinates ) << endl;
      }
}

template< typename Real,
          typename Index >
   template< typename UserData,
             typename EntitiesProcessor >
void
tnlTraverser< tnlGrid< 3, Real, tnlHost, Index >, 2 >::
processInteriorEntities( const GridType& grid,
                         UserData& userData ) const
{
   /****
    * Traversing interior faces
    */
   CoordinatesType coordinates;
   const IndexType& xSize = grid.getDimensions().x();
   const IndexType& ySize = grid.getDimensions().y();
   const IndexType& zSize = grid.getDimensions().z();

#ifdef HAVE_OPENMP
//#pragma omp parallel for
#endif

   //cout << "< 1, 0, 0 >" << endl;
   for( coordinates.z() = 0; coordinates.z() < zSize; coordinates.z() ++ )
      for( coordinates.y() = 0; coordinates.y() < ySize; coordinates.y() ++ )
         for( coordinates.x() = 1; coordinates.x() < xSize; coordinates.x() ++ )
         {
            typedef typename GridType::template Face< 1, 0, 0 > FaceTopology;
            const IndexType index = grid.template getFaceIndex< 1, 0, 0 >( coordinates );
            EntitiesProcessor::template processEntity< FaceTopology >( grid, userData, index, coordinates );
            //cout << "Interior face coordinates = " << coordinates << " index = " << grid.template getFaceIndex< 1, 0 >( coordinates ) << endl;
         }

   //cout << "<  0, 1, 0 >" << endl;
   for( coordinates.z() = 0; coordinates.z() < zSize; coordinates.z() ++ )
      for( coordinates.y() = 1; coordinates.y() < ySize; coordinates.y() ++ )
         for( coordinates.x() = 0; coordinates.x() < xSize; coordinates.x() ++ )
         {
            typedef typename GridType::template Face< 0, 1, 0 > FaceTopology;
            const IndexType index = grid.template getFaceIndex< 0, 1, 0 >( coordinates );
            EntitiesProcessor::template processEntity< FaceTopology >( grid, userData, index, coordinates );
            //cout << "Interior face coordinates = " << coordinates << " index = " << grid.template getFaceIndex< 0, 1 >( coordinates ) << endl;
         }
   
   //cout << "<  0, 1, 0 >" << endl;
   for( coordinates.z() = 1; coordinates.z() < zSize; coordinates.z() ++ )
      for( coordinates.y() = 0; coordinates.y() < ySize; coordinates.y() ++ )
         for( coordinates.x() = 0; coordinates.x() < xSize; coordinates.x() ++ )
         {
            typedef typename GridType::template Face< 0, 0, 1 > FaceTopology;
            const IndexType index = grid.template getFaceIndex< 0, 0, 1 >( coordinates );
            EntitiesProcessor::template processEntity< FaceTopology >( grid, userData, index, coordinates );
            //cout << "Interior face coordinates = " << coordinates << " index = " << grid.template getFaceIndex< 0, 1 >( coordinates ) << endl;
         }   
}


template< typename Real,
          typename Index >
   template< typename UserData,
             typename EntitiesProcessor >
void
tnlTraverser< tnlGrid< 3, Real, tnlHost, Index >, 1 >::
processBoundaryEntities( const GridType& grid,
                         UserData& userData ) const
{
   /****
    * Traversing boundary edges
    */
   CoordinatesType coordinates;
   const IndexType& xSize = grid.getDimensions().x();
   const IndexType& ySize = grid.getDimensions().y();
   const IndexType& zSize = grid.getDimensions().z();

   for( coordinates.y() = 0; coordinates.y() <= ySize; coordinates.y() ++ )
      for( coordinates.x() = 0; coordinates.x() < xSize; coordinates.x() ++ )      
      {
         typedef typename GridType::template Edge< 1, 0, 0 > FaceTopology;
         coordinates.z() = 0;
         EntitiesProcessor::template processEntity< FaceTopology >( grid, userData, grid.template getEdgeIndex< 1, 0, 0 >( coordinates ), coordinates );        
         coordinates.z() = zSize;
         EntitiesProcessor::template processEntity< FaceTopology >( grid, userData, grid.template getEdgeIndex< 1, 0, 0 >( coordinates ), coordinates );
      }
   for( coordinates.y() = 0; coordinates.y() < ySize; coordinates.y() ++ )
      for( coordinates.x() = 0; coordinates.x() <= xSize; coordinates.x() ++ )      
      {
         typedef typename GridType::template Edge< 0, 1, 0 > EdgeTopology;
         coordinates.z() = 0;
         EntitiesProcessor::template processEntity< EdgeTopology >( grid, userData, grid.template getEdgeIndex< 0, 1, 0 >( coordinates ), coordinates );
         coordinates.z() = zSize;
         EntitiesProcessor::template processEntity< EdgeTopology >( grid, userData, grid.template getEdgeIndex< 0, 1, 0 >( coordinates ), coordinates );
      }
     
   for( coordinates.z() = 0; coordinates.z() <= zSize; coordinates.z() ++ )
      for( coordinates.x() = 0; coordinates.x() < xSize; coordinates.x() ++ )
      {
         typedef typename GridType::template Edge< 1, 0, 0 > EdgeTopology;
         coordinates.y() = 0;
         EntitiesProcessor::template processEntity< EdgeTopology >( grid, userData, grid.template getEdgeIndex< 1, 0, 0 >( coordinates ), coordinates );
         coordinates.y() = ySize;
         EntitiesProcessor::template processEntity< EdgeTopology >( grid, userData, grid.template getEdgeIndex< 1, 0, 0 >( coordinates ), coordinates );
      }
   for( coordinates.z() = 0; coordinates.z() < zSize; coordinates.z() ++ )
      for( coordinates.x() = 0; coordinates.x() <= xSize; coordinates.x() ++ )
      {
         typedef typename GridType::template Edge< 0, 0, 1 > EdgeTopology;
         coordinates.y() = 0;
         EntitiesProcessor::template processEntity< EdgeTopology >( grid, userData, grid.template getFaceIndex< 0, 0, 1 >( coordinates ), coordinates );
         coordinates.y() = ySize;
         EntitiesProcessor::template processEntity< EdgeTopology >( grid, userData, grid.template getFaceIndex< 0, 0, 1 >( coordinates ), coordinates );
      }

  
   for( coordinates.z() = 0; coordinates.z() <= zSize; coordinates.z() ++ )
      for( coordinates.y() = 0; coordinates.y() < ySize; coordinates.y() ++ )
      {
         typedef typename GridType::template Edge< 0, 1, 0 > EdgeTopology;
         coordinates.x() = 0;
         EntitiesProcessor::template processEntity< EdgeTopology >( grid, userData, grid.template getEdgeIndex< 0, 1, 0 >( coordinates ), coordinates );
         coordinates.x() = xSize;
         EntitiesProcessor::template processEntity< EdgeTopology >( grid, userData, grid.template getEdgeIndex< 0, 1, 0 >( coordinates ), coordinates );
      }
   for( coordinates.z() = 0; coordinates.z() < zSize; coordinates.z() ++ )
      for( coordinates.y() = 0; coordinates.y() <= ySize; coordinates.y() ++ )
      {
         typedef typename GridType::template Edge< 0, 0, 1 > EdgeTopology;
         coordinates.x() = 0;
         EntitiesProcessor::template processEntity< EdgeTopology >( grid, userData, grid.template getEdgeIndex< 0, 0, 1 >( coordinates ), coordinates );
         coordinates.x() = xSize;
         EntitiesProcessor::template processEntity< EdgeTopology >( grid, userData, grid.template getEdgeIndex< 0, 0, 1 >( coordinates ), coordinates );
      }
}

template< typename Real,
          typename Index >
   template< typename UserData,
             typename EntitiesProcessor >
void
tnlTraverser< tnlGrid< 3, Real, tnlHost, Index >, 1 >::
processInteriorEntities( const GridType& grid,
                         UserData& userData ) const
{
   /****
    * Traversing interior edges
    */
   CoordinatesType coordinates;
   const IndexType& xSize = grid.getDimensions().x();
   const IndexType& ySize = grid.getDimensions().y();
   const IndexType& zSize = grid.getDimensions().z();

#ifdef HAVE_OPENMP
//#pragma omp parallel for
#endif

   //cout << "< 1, 0, 0 >" << endl;
   for( coordinates.z() = 1; coordinates.z() < zSize; coordinates.z() ++ )
      for( coordinates.y() = 1; coordinates.y() < ySize; coordinates.y() ++ )
         for( coordinates.x() = 0; coordinates.x() < xSize; coordinates.x() ++ )
         {
            typedef typename GridType::template Edge< 1, 0, 0 > EdgeTopology;
            const IndexType index = grid.template getEdgeIndex< 1, 0, 0 >( coordinates );
            EntitiesProcessor::template processEntity< EdgeTopology >( grid, userData, index, coordinates );
         }

   //cout << "<  0, 1, 0 >" << endl;
   for( coordinates.z() = 1; coordinates.z() < zSize; coordinates.z() ++ )
      for( coordinates.y() = 0; coordinates.y() < ySize; coordinates.y() ++ )
         for( coordinates.x() = 1; coordinates.x() < xSize; coordinates.x() ++ )
         {
            typedef typename GridType::template Edge< 0, 1, 0 > EdgeTopology;
            const IndexType index = grid.template getEdgeIndex< 0, 1, 0 >( coordinates );
            EntitiesProcessor::template processEntity< EdgeTopology >( grid, userData, index, coordinates );
         }
   
   //cout << "<  0, 1, 0 >" << endl;
   for( coordinates.z() = 0; coordinates.z() < zSize; coordinates.z() ++ )
      for( coordinates.y() = 1; coordinates.y() < ySize; coordinates.y() ++ )
         for( coordinates.x() = 1; coordinates.x() < xSize; coordinates.x() ++ )
         {
            typedef typename GridType::template Edge< 1, 0, 0 > EdgeTopology;
            const IndexType index = grid.template getEdgeIndex< 1, 0, 0 >( coordinates );
            EntitiesProcessor::template processEntity< EdgeTopology >( grid, userData, index, coordinates );
         }   
}


template< typename Real,
          typename Index >
   template< typename UserData,
             typename EntitiesProcessor >
void
tnlTraverser< tnlGrid< 3, Real, tnlHost, Index >, 0 >::
processBoundaryEntities( const GridType& grid,
                         UserData& userData ) const
{
   /****
    * Traversing boundary vertices
    */
   CoordinatesType coordinates;
   const IndexType& xSize = grid.getDimensions().x();
   const IndexType& ySize = grid.getDimensions().y();
   const IndexType& zSize = grid.getDimensions().z();
   typedef typename GridType::Vertex VertexTopology;
   
   for( coordinates.y() = 0; coordinates.y() <= ySize; coordinates.y() ++ )
      for( coordinates.x() = 0; coordinates.x() <= xSize; coordinates.x() ++ )
      {
         coordinates.z() = 0;
         EntitiesProcessor::template processEntity< VertexTopology >( grid, userData, grid.getVertexIndex( coordinates ), coordinates );
         coordinates.z() = zSize;
         EntitiesProcessor::template processEntity< VertexTopology >( grid, userData, grid.getVertexIndex( coordinates ), coordinates );
      }

   for( coordinates.z() = 0; coordinates.z() <= zSize; coordinates.z() ++ )
      for( coordinates.x() = 0; coordinates.x() <= xSize; coordinates.x() ++ )
      {
         coordinates.y() = 0;
         EntitiesProcessor::template processEntity< VertexTopology >( grid, userData, grid.getVertexIndex( coordinates ), coordinates );
         coordinates.y() = ySize;
         EntitiesProcessor::template processEntity< VertexTopology >( grid, userData, grid.getVertexIndex( coordinates ), coordinates );
      }

   for( coordinates.z() = 0; coordinates.z() <= zSize; coordinates.z() ++ )
      for( coordinates.y() = 0; coordinates.y() <= ySize; coordinates.y() ++ )
      {
         coordinates.x() = 0;
         EntitiesProcessor::template processEntity< VertexTopology >( grid, userData, grid.getVertexIndex( coordinates ), coordinates );
         coordinates.x() = xSize;
         EntitiesProcessor::template processEntity< VertexTopology >( grid, userData, grid.getVertexIndex( coordinates ), coordinates );
      }
}

template< typename Real,
          typename Index >
   template< typename UserData,
             typename EntitiesProcessor >
void
tnlTraverser< tnlGrid< 3, Real, tnlHost, Index >, 0 >::
processInteriorEntities( const GridType& grid,
                         UserData& userData ) const
{
   /****
    * Traversing interior vertices
    */
   CoordinatesType coordinates;
   const IndexType& xSize = grid.getDimensions().x();
   const IndexType& ySize = grid.getDimensions().y();
   const IndexType& zSize = grid.getDimensions().z();
   typedef typename GridType::Vertex VertexTopology;

#ifdef HAVE_OPENMP
//#pragma omp parallel for
#endif
   for( coordinates.z() = 1; coordinates.z() < zSize; coordinates.z() ++ )
      for( coordinates.y() = 1; coordinates.y() < ySize; coordinates.y() ++ )
         for( coordinates.x() = 1; coordinates.x() < xSize; coordinates.x() ++ )
         {
            const IndexType index = grid.getVertexIndex( coordinates );
            EntitiesProcessor::template processEntity< VertexTopology >( grid, userData, index, coordinates );
         }
}


/***
 *
 *    CUDA Specializations
 *
 */
#ifdef HAVE_CUDA
template< typename Real,
          typename Index,
          typename UserData,
          typename EntitiesProcessor >
__global__ void tnlTraverserGrid3DBoundaryCells( const tnlGrid< 3, Real, tnlCuda, Index >* grid,
                                                 UserData* userData,
                                                 const Index gridXIdx,
                                                 const Index gridYIdx,
                                                 const Index gridZIdx )
{
   typedef Real RealType;
   typedef Index IndexType;
   typedef tnlGrid< 3, Real, tnlCuda, Index > GridType;
   typedef typename GridType::CoordinatesType CoordinatesType;
   typedef typename GridType::Cell CellTopology;

   const IndexType& xSize = grid->getDimensions().x();
   const IndexType& ySize = grid->getDimensions().y();
   const IndexType& zSize = grid->getDimensions().z();

   CoordinatesType cellCoordinates( ( gridXIdx * tnlCuda::getMaxGridSize() + blockIdx.x ) * blockDim.x + threadIdx.x,
                                    ( gridYIdx * tnlCuda::getMaxGridSize() + blockIdx.y ) * blockDim.y + threadIdx.y,
                                    ( gridZIdx * tnlCuda::getMaxGridSize() + blockIdx.z ) * blockDim.z + threadIdx.z );

   if( cellCoordinates.x() < grid->getDimensions().x() &&
       cellCoordinates.y() < grid->getDimensions().y() &&
       cellCoordinates.z() < grid->getDimensions().z() )
   {
      if( grid->isBoundaryCell( cellCoordinates ) )
      {
         //printf( "Processing boundary conditions at %d %d \n", cellCoordinates.x(), cellCoordinates.y() );
         EntitiesProcessor::template processEntity< CellTopology >
            ( *grid,
              *userData,
              grid->getCellIndex( cellCoordinates ),
              cellCoordinates );
      }
   }
}

template< typename Real,
          typename Index,
          typename UserData,
          typename EntitiesProcessor >
__global__ void tnlTraverserGrid3DInteriorCells( const tnlGrid< 3, Real, tnlCuda, Index >* grid,
                                                 UserData* userData,
                                                 const Index gridXIdx,
                                                 const Index gridYIdx,
                                                 const Index gridZIdx )
{
   typedef Real RealType;
   typedef Index IndexType;
   typedef tnlGrid< 3, Real, tnlCuda, Index > GridType;
   typedef typename GridType::CoordinatesType CoordinatesType;
   typedef typename GridType::Cell CellTopology;

   const IndexType& xSize = grid->getDimensions().x();
   const IndexType& ySize = grid->getDimensions().y();

   CoordinatesType cellCoordinates( ( gridXIdx * tnlCuda::getMaxGridSize() + blockIdx.x ) * blockDim.x + threadIdx.x,
                                    ( gridYIdx * tnlCuda::getMaxGridSize() + blockIdx.y ) * blockDim.y + threadIdx.y,
                                    ( gridZIdx * tnlCuda::getMaxGridSize() + blockIdx.z ) * blockDim.z + threadIdx.z );

   if( cellCoordinates.x() < grid->getDimensions().x() &&
       cellCoordinates.y() < grid->getDimensions().y() &&
       cellCoordinates.z() < grid->getDimensions().z())
   {
      if( ! grid->isBoundaryCell( cellCoordinates ) )
      {
         //printf( "Processing interior conditions at %d %d \n", cellCoordinates.x(), cellCoordinates.y() );
         EntitiesProcessor::template processEntity< CellTopology >
            ( *grid,
              *userData,
              grid->getCellIndex( cellCoordinates ),
              cellCoordinates );
      }
   }
}

template< typename Real,
          typename Index,
          typename UserData,
          typename EntitiesProcessor >
__global__ void tnlTraverserGrid3DBoundaryVertices( const tnlGrid< 3, Real, tnlCuda, Index >* grid,
                                                    UserData* userData,
                                                    const Index gridXIdx,
                                                    const Index gridYIdx,
                                                    const Index gridZIdx )
{
   typedef Real RealType;
   typedef Index IndexType;
   typedef tnlGrid< 3, Real, tnlCuda, Index > GridType;
   typedef typename GridType::CoordinatesType CoordinatesType;
   typedef typename GrigType::Vertex VertexTopology;
   
   const IndexType& xSize = grid->getDimensions().x();
   const IndexType& ySize = grid->getDimensions().y();
   const IndexType& zSize = grid->getDimensions().z();

   CoordinatesType vertexCoordinates( ( gridXIdx * tnlCuda::getMaxGridSize() + blockIdx.x ) * blockDim.x + threadIdx.x,
                                      ( gridYIdx * tnlCuda::getMaxGridSize() + blockIdx.y ) * blockDim.y + threadIdx.y,
                                      ( gridZIdx * tnlCuda::getMaxGridSize() + blockIdx.z ) * blockDim.z + threadIdx.z );

   if( vertexCoordinates.x() < grid->getDimensions().x() &&
       vertexCoordinates.y() < grid->getDimensions().y() &&
       vertexCoordinates.z() < grid->getDimensions().z() )
   {
      if( grid->isBoundaryVertex( vertexCoordinates ) )
      {
         EntitiesProcessor::template processEntity< VertexTopology >
         ( *grid,
           *userData,
           grid->getVertexIndex( vertexCoordinates ),
           vertexCoordinates );
      }
   }
}

template< typename Real,
          typename Index,
          typename UserData,
          typename EntitiesProcessor,
          int nx,
          int ny,
          int nz >
__global__ void tnlTraverserGrid3DBoundaryFaces( const tnlGrid< 3, Real, tnlCuda, Index >* grid,
                                                 UserData* userData,
                                                 const Index gridXIdx,
                                                 const Index gridYIdx,
                                                 const Index gridZIdx )
{
   typedef Real RealType;
   typedef Index IndexType;
   typedef tnlGrid< 3, Real, tnlCuda, Index > GridType;
   typedef typename GridType::CoordinatesType CoordinatesType;
   typedef typename GridType::template Face< nx, ny, nz > FaceTopology;

   const IndexType& xSize = grid->getDimensions().x();
   const IndexType& ySize = grid->getDimensions().y();
   const IndexType& zSize = grid->getDimensions().z();

   CoordinatesType faceCoordinates( ( gridXIdx * tnlCuda::getMaxGridSize() + blockIdx.x ) * blockDim.x + threadIdx.x,
                                    ( gridYIdx * tnlCuda::getMaxGridSize() + blockIdx.y ) * blockDim.y + threadIdx.y,
                                    ( gridZIdx * tnlCuda::getMaxGridSize() + blockIdx.z ) * blockDim.z + threadIdx.z    );

   if( faceCoordinates.x() < grid->getDimensions().x() + nx &&
       faceCoordinates.y() < grid->getDimensions().y() + ny &&
       faceCoordinates.z() < grid->getDimensions().z() + nz )
   {
      if( grid->template isBoundaryFace< nx, ny, nz >( faceCoordinates ) )
      {
         //printf( "Processing boundary conditions at %d %d \n", cellCoordinates.x(), cellCoordinates.y() );
         EntitiesProcessor::template processEntity< FaceTopology >
            ( *grid,
              *userData,
              grid->template getFaceIndex< nx, ny, nz >( faceCoordinates ),
              faceCoordinates );
      }
   }
}

template< typename Real,
          typename Index,
          typename UserData,
          typename EntitiesProcessor,
          int nx,
          int ny,
          int nz >
__global__ void tnlTraverserGrid3DInteriorFaces( const tnlGrid< 3, Real, tnlCuda, Index >* grid,
                                                 UserData* userData,
                                                 const Index gridXIdx,
                                                 const Index gridYIdx,
                                                 cinst Index gridZIdx )
{
   typedef Real RealType;
   typedef Index IndexType;
   typedef tnlGrid< 3, Real, tnlCuda, Index > GridType;
   typedef typename GridType::CoordinatesType CoordinatesType;
   typedef typename GridType::template Face< nx, ny, nz > FaceTopology;

   const IndexType& xSize = grid->getDimensions().x();
   const IndexType& ySize = grid->getDimensions().y();
   const IndexType& zSize = grid->getDimensions().z();

   CoordinatesType faceCoordinates( ( gridXIdx * tnlCuda::getMaxGridSize() + blockIdx.x ) * blockDim.x + threadIdx.x,
                                    ( gridYIdx * tnlCuda::getMaxGridSize() + blockIdx.y ) * blockDim.y + threadIdx.y,
                                    ( gridZIdx * tnlCuda::getMaxGridSize() + blockIdx.z ) * blockDim.z + threadIdx.z );

   if( faceCoordinates.x() < grid->getDimensions().x() + nx &&
       faceCoordinates.y() < grid->getDimensions().y() + ny && 
       faceCoordinates.z() < grid->getDimensions().z() + nz )
   {
      if( ! grid->template isBoundaryFace< nx, ny, nz >( faceCoordinates ) )
      {
         //printf( "Processing interior conditions at %d %d \n", cellCoordinates.x(), cellCoordinates.y() );
         EntitiesProcessor::template processEntity< FaceTopology >
            ( *grid,
              *userData,
              grid->template getFaceIndex< nx, ny, nz >( faceCoordinates ),
              faceCoordinates );
      }
   }
}

template< typename Real,
          typename Index,
          typename UserData,
          typename EntitiesProcessor,
          int dx,
          int dy,
          int dz >
__global__ void tnlTraverserGrid3DBoundaryEdges( const tnlGrid< 3, Real, tnlCuda, Index >* grid,
                                                  UserData* userData,
                                                  const Index gridXIdx,
                                                  const Index gridYIdx,
                                                  const Index gridZIdx )
{
   typedef Real RealType;
   typedef Index IndexType;
   typedef tnlGrid< 3, Real, tnlCuda, Index > GridType;
   typedef typename GridType::CoordinatesType CoordinatesType;
   typedef typename GridType::template Edge< dx, dy, dz > EdgeTopology;

   const IndexType& xSize = grid->getDimensions().x();
   const IndexType& ySize = grid->getDimensions().y();
   const IndexType& zSize = grid->getDimensions().z();

   CoordinatesType edgeCoordinates( ( gridXIdx * tnlCuda::getMaxGridSize() + blockIdx.x ) * blockDim.x + threadIdx.x,
                                    ( gridYIdx * tnlCuda::getMaxGridSize() + blockIdx.y ) * blockDim.y + threadIdx.y,
                                    ( gridZIdx * tnlCuda::getMaxGridSize() + blockIdx.z ) * blockDim.z + threadIdx.z    );

   if( edgeCoordinates.x() < grid->getDimensions().x() + !nx &&
       faceCoordinates.y() < grid->getDimensions().y() + !ny &&
       faceCoordinates.z() < grid->getDimensions().z() + !nz )
   {
      if( grid->template isBoundaryEdge< dx, dy, dz >( edgeCoordinates ) )
      {
         //printf( "Processing boundary conditions at %d %d \n", cellCoordinates.x(), cellCoordinates.y() );
         EntitiesProcessor::template processEntity< EdgeTopology >
            ( *grid,
              *userData,
              grid->template getFaceIndex< dx, dy, dz >( edgeCoordinates ),
              edgeCoordinates );
      }
   }
}

template< typename Real,
          typename Index,
          typename UserData,
          typename EntitiesProcessor,
          int dx,
          int dy,
          int dz >
__global__ void tnlTraverserGrid3DInteriorEdges( const tnlGrid< 3, Real, tnlCuda, Index >* grid,
                                                 UserData* userData,
                                                 const Index gridXIdx,
                                                 const Index gridYIdx,
                                                 cinst Index gridZIdx )
{
   typedef Real RealType;
   typedef Index IndexType;
   typedef tnlGrid< 3, Real, tnlCuda, Index > GridType;
   typedef typename GridType::CoordinatesType CoordinatesType;
   typedef typename GridType::template Face< dx, dy, dz > EdgeTopology;

   const IndexType& xSize = grid->getDimensions().x();
   const IndexType& ySize = grid->getDimensions().y();
   const IndexType& zSize = grid->getDimensions().z();

   CoordinatesType edgeCoordinates( ( gridXIdx * tnlCuda::getMaxGridSize() + blockIdx.x ) * blockDim.x + threadIdx.x,
                                    ( gridYIdx * tnlCuda::getMaxGridSize() + blockIdx.y ) * blockDim.y + threadIdx.y,
                                    ( gridZIdx * tnlCuda::getMaxGridSize() + blockIdx.z ) * blockDim.z + threadIdx.z );

   if( edgeCoordinates.x() < grid->getDimensions().x() + !dx &&
       edgeCoordinates.y() < grid->getDimensions().y() + !dy && 
       edgeCoordinates.z() < grid->getDimensions().z() + !dz )
   {
      if( ! grid->template isBoundaryEdge< dx, dy, dz >( edgeCoordinates ) )
      {
         //printf( "Processing interior conditions at %d %d \n", cellCoordinates.x(), cellCoordinates.y() );
         EntitiesProcessor::template processEntity< EdgeTopology >
            ( *grid,
              *userData,
              grid->template getEdgeIndex< dx, dy, dz >( edgeCoordinates ),
              edgeCoordinates );
      }
   }
}


template< typename Real,
          typename Index,
          typename UserData,
          typename EntitiesProcessor >
__global__ void tnlTraverserGrid3DInteriorVertices( const tnlGrid< 3, Real, tnlCuda, Index >* grid,
                                                    UserData* userData,
                                                    const Index gridXIdx,
                                                    const Index gridYIdx,
                                                    const Index gridZIdx )
{
   typedef Real RealType;
   typedef Index IndexType;
   typedef tnlGrid< 3, Real, tnlCuda, Index > GridType;
   typedef typename GridType::CoordinatesType CoordinatesType;

   const IndexType& xSize = grid->getDimensions().x();
   const IndexType& ySize = grid->getDimensions().y();
   const IndexType& zSize = grid->getDimensions().z();

   CoordinatesType vertexCoordinates( ( gridXIdx * tnlCuda::getMaxGridSize() + blockIdx.x ) * blockDim.x + threadIdx.x,
                                      ( gridYIdx * tnlCuda::getMaxGridSize() + blockIdx.y ) * blockDim.y + threadIdx.y,
                                      ( gridZIdx * tnlCuda::getMaxGridSize() + blockIdx.z ) * blockDim.z + threadIdx.z );

   if( vertexCoordinates.x() < grid->getDimensions().x() &&
       vertexCoordinates.y() < grid->getDimensions().y() &&
       vertexCoordinates.z() < grid->getDimensions().z() )
   {
      if( ! grid->isBoundaryVertex( vertexCoordinates ) )
      {
         EntitiesProcessor::processVertex( *grid,
                                           *userData,
                                           grid->getVertexIndex( vertexCoordinates ),
                                           vertexCoordinates );
      }
   }
}


#endif


template< typename Real,
          typename Index >
   template< typename UserData,
             typename EntitiesProcessor >
void
tnlTraverser< tnlGrid< 3, Real, tnlCuda, Index >, 3 >::
processBoundaryEntities( const GridType& grid,
                         UserData& userData ) const
{
#ifdef HAVE_CUDA
   /****
    * Boundary cells
    */
   GridType* kernelGrid = tnlCuda::passToDevice( grid );
   UserData* kernelUserData = tnlCuda::passToDevice( userData );

   dim3 cudaBlockSize( 8, 8, 4 );
   dim3 cudaBlocks;
   cudaBlocks.x = tnlCuda::getNumberOfBlocks( grid.getDimensions().x(), cudaBlockSize.x );
   cudaBlocks.y = tnlCuda::getNumberOfBlocks( grid.getDimensions().y(), cudaBlockSize.y );
   cudaBlocks.z = tnlCuda::getNumberOfBlocks( grid.getDimensions().z(), cudaBlockSize.z );
   const IndexType cudaXGrids = tnlCuda::getNumberOfGrids( cudaBlocks.x );
   const IndexType cudaYGrids = tnlCuda::getNumberOfGrids( cudaBlocks.y );
   const IndexType cudaZGrids = tnlCuda::getNumberOfGrids( cudaBlocks.z );

   for( IndexType gridXIdx = 0; gridXIdx < cudaXGrids; gridXIdx ++ )
      for( IndexType gridYIdx = 0; gridYIdx < cudaYGrids; gridYIdx ++ )
         for( IndexType gridZIdx = 0; gridZIdx < cudaZGrids; gridZIdx ++ )
         {
            tnlTraverserGrid3DBoundaryCells< Real, Index, UserData, EntitiesProcessor >
                                           <<< cudaBlocks, cudaBlockSize >>>
                                          ( kernelGrid,
                                            kernelUserData,
                                            gridXIdx,
                                            gridYIdx,
                                            gridZIdx );
         }
   cudaThreadSynchronize();
   checkCudaDevice;
#endif
}

template< typename Real,
          typename Index >
   template< typename UserData,
             typename EntitiesProcessor >
void
tnlTraverser< tnlGrid< 3, Real, tnlCuda, Index >, 3 >::
processInteriorEntities( const GridType& grid,
                         UserData& userData ) const
{
   #ifdef HAVE_CUDA

   /****
    * Interior cells
    */
   GridType* kernelGrid = tnlCuda::passToDevice( grid );
   UserData* kernelUserData = tnlCuda::passToDevice( userData );

   dim3 cudaBlockSize( 8, 8, 4 );
   dim3 cudaBlocks;
   cudaBlocks.x = tnlCuda::getNumberOfBlocks( grid.getDimensions().x(), cudaBlockSize.x );
   cudaBlocks.y = tnlCuda::getNumberOfBlocks( grid.getDimensions().y(), cudaBlockSize.y );
   cudaBlocks.z = tnlCuda::getNumberOfBlocks( grid.getDimensions().z(), cudaBlockSize.z );
   const IndexType cudaXGrids = tnlCuda::getNumberOfGrids( cudaBlocks.x );
   const IndexType cudaYGrids = tnlCuda::getNumberOfGrids( cudaBlocks.y );
   const IndexType cudaZGrids = tnlCuda::getNumberOfGrids( cudaBlocks.z );

   for( IndexType gridXIdx = 0; gridXIdx < cudaXGrids; gridXIdx ++ )
      for( IndexType gridYIdx = 0; gridYIdx < cudaYGrids; gridYIdx ++ )
         for( IndexType gridZIdx = 0; gridZIdx < cudaZGrids; gridZIdx ++ )
         {
            tnlTraverserGrid3DInteriorCells< Real, Index, UserData, EntitiesProcessor >
                                           <<< cudaBlocks, cudaBlockSize >>>
                                          ( kernelGrid,
                                            kernelUserData,
                                            gridXIdx,
                                            gridYIdx,
                                            gridZIdx );
         }
   checkCudaDevice;
   tnlCuda::freeFromDevice( kernelGrid );
   tnlCuda::freeFromDevice( kernelUserData );
#endif
}


template< typename Real,
          typename Index >
   template< typename UserData,
             typename EntitiesProcessor >
void
tnlTraverser< tnlGrid< 3, Real, tnlCuda, Index >, 2 >::
processBoundaryEntities( const GridType& grid,
                         UserData& userData ) const
{
#ifdef HAVE_CUDA
   /****
    * Boundary faces
    */
   GridType* kernelGrid = tnlCuda::passToDevice( grid );
   UserData* kernelUserData = tnlCuda::passToDevice( userData );

   dim3 cudaBlockSize( 8, 8, 4 );
   dim3 cudaBlocks;
   IndexType cudaXGrids, cudaYGrids, cudaZGrids;

   /****
    * < 1, 0, 0 > faces
    */
   cudaBlocks.x = tnlCuda::getNumberOfBlocks( grid.getDimensions().x() + 1, cudaBlockSize.x );
   cudaBlocks.y = tnlCuda::getNumberOfBlocks( grid.getDimensions().y(), cudaBlockSize.y );
   cudaBlocks.z = tnlCuda::getNumberOfBlocks( grid.getDimensions().z(), cudaBlockSize.z );
   cudaXGrids = tnlCuda::getNumberOfGrids( cudaBlocks.x );
   cudaYGrids = tnlCuda::getNumberOfGrids( cudaBlocks.y );
   cudaZGrids = tnlCuda::getNumberOfGrids( cudaBlocks.z );
   for( IndexType gridXIdx = 0; gridXIdx < cudaXGrids; gridXIdx ++ )
      for( IndexType gridYIdx = 0; gridYIdx < cudaYGrids; gridYIdx ++ )
         for( IndexType gridZIdx = 0; gridZIdx < cudaZGrids; gridZIdx ++ )
         {
            tnlTraverserGrid3DBoundaryFaces< Real, Index, UserData, EntitiesProcessor, 1, 0, 0 >
                                           <<< cudaBlocks, cudaBlockSize >>>
                                          ( kernelGrid,
                                            kernelUserData,
                                            gridXIdx,
                                            gridYIdx,
                                            gridZIdx );
            checkCudaDevice;
         }
   cudaThreadSynchronize();
   
   /****
    * < 0, 1, 0 > faces
    */
   cudaBlocks.x = tnlCuda::getNumberOfBlocks( grid.getDimensions().x(), cudaBlockSize.x );
   cudaBlocks.y = tnlCuda::getNumberOfBlocks( grid.getDimensions().y() + 1, cudaBlockSize.y );
   cudaBlocks.z = tnlCuda::getNumberOfBlocks( grid.getDimensions().z(), cudaBlockSize.z );
   cudaXGrids = tnlCuda::getNumberOfGrids( cudaBlocks.x );
   cudaYGrids = tnlCuda::getNumberOfGrids( cudaBlocks.y );
   cudaZGrids = tnlCuda::getNumberOfGrids( cudaBlocks.z);
   for( IndexType gridXIdx = 0; gridXIdx < cudaXGrids; gridXIdx ++ )
      for( IndexType gridYIdx = 0; gridYIdx < cudaYGrids; gridYIdx ++ )
         for( IndexType gridZIdx = 0; gridZIdx < cudaZGrids; gridZIdx ++ )
         {
            tnlTraverserGrid3DBoundaryFaces< Real, Index, UserData, EntitiesProcessor, 0, 1, 0 >
                                           <<< cudaBlocks, cudaBlockSize >>>
                                          ( kernelGrid,
                                            kernelUserData,
                                            gridXIdx,
                                            gridYIdx,
                                            gridZIdx );
            checkCudaDevice;
         }
   cudaThreadSynchronize();
   
   /****
    * < 0, 0, 1 > faces
    */
   cudaBlocks.x = tnlCuda::getNumberOfBlocks( grid.getDimensions().x(), cudaBlockSize.x );
   cudaBlocks.y = tnlCuda::getNumberOfBlocks( grid.getDimensions().y(), cudaBlockSize.y );
   cudaBlocks.z = tnlCuda::getNumberOfBlocks( grid.getDimensions().z() + 1, cudaBlockSize.z );
   cudaXGrids = tnlCuda::getNumberOfGrids( cudaBlocks.x );
   cudaYGrids = tnlCuda::getNumberOfGrids( cudaBlocks.y );
   cudaZGrids = tnlCuda::getNumberOfGrids( cudaBlocks.z);
   for( IndexType gridXIdx = 0; gridXIdx < cudaXGrids; gridXIdx ++ )
      for( IndexType gridYIdx = 0; gridYIdx < cudaYGrids; gridYIdx ++ )
         for( IndexType gridZIdx = 0; gridZIdx < cudaZGrids; gridZIdx ++ )
         {
            tnlTraverserGrid3DBoundaryFaces< Real, Index, UserData, EntitiesProcessor, 0, 0, 1 >
                                           <<< cudaBlocks, cudaBlockSize >>>
                                          ( kernelGrid,
                                            kernelUserData,
                                            gridXIdx,
                                            gridYIdx,
                                            gridZIdx );
            checkCudaDevice;
         }
   cudaThreadSynchronize();
#endif
}

template< typename Real,
          typename Index >
   template< typename UserData,
             typename EntitiesProcessor >
void
tnlTraverser< tnlGrid< 3, Real, tnlCuda, Index >, 2 >::
processInteriorEntities( const GridType& grid,
                         UserData& userData ) const
{
#ifdef HAVE_CUDA
   /****
    * Traversing interior faces
    */
   GridType* kernelGrid = tnlCuda::passToDevice( grid );
   UserData* kernelUserData = tnlCuda::passToDevice( userData );

   dim3 cudaBlockSize( 8, 8, 4 );
   dim3 cudaBlocks;
   IndexType cudaXGrids, cudaYGrids, cudaZGrids;

   /****
    * < 1, 0, 0 > faces
    */
   cudaBlocks.x = tnlCuda::getNumberOfBlocks( grid.getDimensions().x() + 1, cudaBlockSize.x );
   cudaBlocks.y = tnlCuda::getNumberOfBlocks( grid.getDimensions().y(), cudaBlockSize.y );
   cudaBlocks.z = tnlCuda::getNumberOfBlocks( grid.getDimensions().z(), cudaBlockSize.z );
   cudaXGrids = tnlCuda::getNumberOfGrids( cudaBlocks.x );
   cudaYGrids = tnlCuda::getNumberOfGrids( cudaBlocks.y );
   cudaZGrids = tnlCuda::getNumberOfGrids( cudaBlocks.z );
   for( IndexType gridXIdx = 0; gridXIdx < cudaXGrids; gridXIdx ++ )
      for( IndexType gridYIdx = 0; gridYIdx < cudaYGrids; gridYIdx ++ )
         for( IndexType gridZIdx = 0; gridZIdx < cudaZGrids; gridZIdx ++ )
         {
            tnlTraverserGrid3DInteriorFaces< Real, Index, UserData, EntitiesProcessor, 1, 0, 0 >
                                           <<< cudaBlocks, cudaBlockSize >>>
                                          ( kernelGrid,
                                            kernelUserData,
                                            gridXIdx,
                                            gridYIdx,
                                            gridZIdx );
            checkCudaDevice;
         }
   cudaThreadSynchronize();
   
   /****
    * < 0, 1, 0 > faces
    */
   cudaBlocks.x = tnlCuda::getNumberOfBlocks( grid.getDimensions().x(), cudaBlockSize.x );
   cudaBlocks.y = tnlCuda::getNumberOfBlocks( grid.getDimensions().y() + 1, cudaBlockSize.y );
   cudaBlocks.z = tnlCuda::getNumberOfBlocks( grid.getDimensions().z(), cudaBlockSize.z );
   cudaXGrids = tnlCuda::getNumberOfGrids( cudaBlocks.x );
   cudaYGrids = tnlCuda::getNumberOfGrids( cudaBlocks.y );
   cudaZGrids = tnlCuda::getNumberOfGrids( cudaBlocks.z);
   for( IndexType gridXIdx = 0; gridXIdx < cudaXGrids; gridXIdx ++ )
      for( IndexType gridYIdx = 0; gridYIdx < cudaYGrids; gridYIdx ++ )
         for( IndexType gridZIdx = 0; gridZIdx < cudaZGrids; gridZIdx ++ )
         {
            tnlTraverserGrid3DInteriorFaces< Real, Index, UserData, EntitiesProcessor, 0, 1, 0 >
                                           <<< cudaBlocks, cudaBlockSize >>>
                                          ( kernelGrid,
                                            kernelUserData,
                                            gridXIdx,
                                            gridYIdx,
                                            gridZIdx );
            checkCudaDevice;
         }
   cudaThreadSynchronize();
   
   /****
    * < 0, 0, 1 > faces
    */
   cudaBlocks.x = tnlCuda::getNumberOfBlocks( grid.getDimensions().x(), cudaBlockSize.x );
   cudaBlocks.y = tnlCuda::getNumberOfBlocks( grid.getDimensions().y(), cudaBlockSize.y );
   cudaBlocks.z = tnlCuda::getNumberOfBlocks( grid.getDimensions().z() + 1, cudaBlockSize.z );
   cudaXGrids = tnlCuda::getNumberOfGrids( cudaBlocks.x );
   cudaYGrids = tnlCuda::getNumberOfGrids( cudaBlocks.y );
   cudaZGrids = tnlCuda::getNumberOfGrids( cudaBlocks.z);
   for( IndexType gridXIdx = 0; gridXIdx < cudaXGrids; gridXIdx ++ )
      for( IndexType gridYIdx = 0; gridYIdx < cudaYGrids; gridYIdx ++ )
         for( IndexType gridZIdx = 0; gridZIdx < cudaZGrids; gridZIdx ++ )
         {
            tnlTraverserGrid3DInteriorFaces< Real, Index, UserData, EntitiesProcessor, 0, 0, 1 >
                                           <<< cudaBlocks, cudaBlockSize >>>
                                          ( kernelGrid,
                                            kernelUserData,
                                            gridXIdx,
                                            gridYIdx,
                                            gridZIdx );
            checkCudaDevice;
         }
   cudaThreadSynchronize();
#endif
}


template< typename Real,
          typename Index >
   template< typename UserData,
             typename EntitiesProcessor >
void
tnlTraverser< tnlGrid< 3, Real, tnlCuda, Index >, 1 >::
processBoundaryEntities( const GridType& grid,
                         UserData& userData ) const
{
#ifdef HAVE_CUDA
   /****
    * Boundary edges
    */
   GridType* kernelGrid = tnlCuda::passToDevice( grid );
   UserData* kernelUserData = tnlCuda::passToDevice( userData );

   dim3 cudaBlockSize( 8, 8, 4 );
   dim3 cudaBlocks;
   IndexType cudaXGrids, cudaYGrids, cudaZGrids;

   /****
    * < 1, 0, 0 > edges
    */
   cudaBlocks.x = tnlCuda::getNumberOfBlocks( grid.getDimensions().x(), cudaBlockSize.x );
   cudaBlocks.y = tnlCuda::getNumberOfBlocks( grid.getDimensions().y() + 1, cudaBlockSize.y );
   cudaBlocks.z = tnlCuda::getNumberOfBlocks( grid.getDimensions().z() + 1, cudaBlockSize.z );
   cudaXGrids = tnlCuda::getNumberOfGrids( cudaBlocks.x );
   cudaYGrids = tnlCuda::getNumberOfGrids( cudaBlocks.y );
   cudaZGrids = tnlCuda::getNumberOfGrids( cudaBlocks.z );
   for( IndexType gridXIdx = 0; gridXIdx < cudaXGrids; gridXIdx ++ )
      for( IndexType gridYIdx = 0; gridYIdx < cudaYGrids; gridYIdx ++ )
         for( IndexType gridZIdx = 0; gridZIdx < cudaZGrids; gridZIdx ++ )
         {
            tnlTraverserGrid3DBoundaryEdges< Real, Index, UserData, EntitiesProcessor, 1, 0, 0 >
                                           <<< cudaBlocks, cudaBlockSize >>>
                                          ( kernelGrid,
                                            kernelUserData,
                                            gridXIdx,
                                            gridYIdx,
                                            gridZIdx );
            checkCudaDevice;
         }
   cudaThreadSynchronize();
   
   /****
    * < 0, 1, 0 > edges
    */
   cudaBlocks.x = tnlCuda::getNumberOfBlocks( grid.getDimensions().x() + 1, cudaBlockSize.x );
   cudaBlocks.y = tnlCuda::getNumberOfBlocks( grid.getDimensions().y(), cudaBlockSize.y );
   cudaBlocks.z = tnlCuda::getNumberOfBlocks( grid.getDimensions().z() + 1, cudaBlockSize.z );
   cudaXGrids = tnlCuda::getNumberOfGrids( cudaBlocks.x );
   cudaYGrids = tnlCuda::getNumberOfGrids( cudaBlocks.y );
   cudaZGrids = tnlCuda::getNumberOfGrids( cudaBlocks.z);
   for( IndexType gridXIdx = 0; gridXIdx < cudaXGrids; gridXIdx ++ )
      for( IndexType gridYIdx = 0; gridYIdx < cudaYGrids; gridYIdx ++ )
         for( IndexType gridZIdx = 0; gridZIdx < cudaZGrids; gridZIdx ++ )
         {
            tnlTraverserGrid3DBoundaryEdges< Real, Index, UserData, EntitiesProcessor, 0, 1, 0 >
                                           <<< cudaBlocks, cudaBlockSize >>>
                                          ( kernelGrid,
                                            kernelUserData,
                                            gridXIdx,
                                            gridYIdx,
                                            gridZIdx );
            checkCudaDevice;
         }
   cudaThreadSynchronize();
   
   /****
    * < 0, 0, 1 > edges
    */
   cudaBlocks.x = tnlCuda::getNumberOfBlocks( grid.getDimensions().x() + 1, cudaBlockSize.x );
   cudaBlocks.y = tnlCuda::getNumberOfBlocks( grid.getDimensions().y() + 1, cudaBlockSize.y );
   cudaBlocks.z = tnlCuda::getNumberOfBlocks( grid.getDimensions().z(), cudaBlockSize.z );
   cudaXGrids = tnlCuda::getNumberOfGrids( cudaBlocks.x );
   cudaYGrids = tnlCuda::getNumberOfGrids( cudaBlocks.y );
   cudaZGrids = tnlCuda::getNumberOfGrids( cudaBlocks.z);
   for( IndexType gridXIdx = 0; gridXIdx < cudaXGrids; gridXIdx ++ )
      for( IndexType gridYIdx = 0; gridYIdx < cudaYGrids; gridYIdx ++ )
         for( IndexType gridZIdx = 0; gridZIdx < cudaZGrids; gridZIdx ++ )
         {
            tnlTraverserGrid3DBoundaryEdges< Real, Index, UserData, EntitiesProcessor, 0, 0, 1 >
                                           <<< cudaBlocks, cudaBlockSize >>>
                                          ( kernelGrid,
                                            kernelUserData,
                                            gridXIdx,
                                            gridYIdx,
                                            gridZIdx );
            checkCudaDevice;
         }
   cudaThreadSynchronize();
#endif
}

template< typename Real,
          typename Index >
   template< typename UserData,
             typename EntitiesProcessor >
void
tnlTraverser< tnlGrid< 3, Real, tnlCuda, Index >, 1 >::
processInteriorEntities( const GridType& grid,
                         UserData& userData ) const
{
#ifdef HAVE_CUDA
   /****
    * Boundary edges
    */
   GridType* kernelGrid = tnlCuda::passToDevice( grid );
   UserData* kernelUserData = tnlCuda::passToDevice( userData );

   dim3 cudaBlockSize( 8, 8, 4 );
   dim3 cudaBlocks;
   IndexType cudaXGrids, cudaYGrids, cudaZGrids;

   /****
    * < 1, 0, 0 > edges
    */
   cudaBlocks.x = tnlCuda::getNumberOfBlocks( grid.getDimensions().x(), cudaBlockSize.x );
   cudaBlocks.y = tnlCuda::getNumberOfBlocks( grid.getDimensions().y() + 1, cudaBlockSize.y );
   cudaBlocks.z = tnlCuda::getNumberOfBlocks( grid.getDimensions().z() + 1, cudaBlockSize.z );
   cudaXGrids = tnlCuda::getNumberOfGrids( cudaBlocks.x );
   cudaYGrids = tnlCuda::getNumberOfGrids( cudaBlocks.y );
   cudaZGrids = tnlCuda::getNumberOfGrids( cudaBlocks.z );
   for( IndexType gridXIdx = 0; gridXIdx < cudaXGrids; gridXIdx ++ )
      for( IndexType gridYIdx = 0; gridYIdx < cudaYGrids; gridYIdx ++ )
         for( IndexType gridZIdx = 0; gridZIdx < cudaZGrids; gridZIdx ++ )
         {
            tnlTraverserGrid3DInteriorEdges< Real, Index, UserData, EntitiesProcessor, 1, 0, 0 >
                                           <<< cudaBlocks, cudaBlockSize >>>
                                          ( kernelGrid,
                                            kernelUserData,
                                            gridXIdx,
                                            gridYIdx,
                                            gridZIdx );
            checkCudaDevice;
         }
   cudaThreadSynchronize();
   
   /****
    * < 0, 1, 0 > edges
    */
   cudaBlocks.x = tnlCuda::getNumberOfBlocks( grid.getDimensions().x() + 1, cudaBlockSize.x );
   cudaBlocks.y = tnlCuda::getNumberOfBlocks( grid.getDimensions().y(), cudaBlockSize.y );
   cudaBlocks.z = tnlCuda::getNumberOfBlocks( grid.getDimensions().z() + 1, cudaBlockSize.z );
   cudaXGrids = tnlCuda::getNumberOfGrids( cudaBlocks.x );
   cudaYGrids = tnlCuda::getNumberOfGrids( cudaBlocks.y );
   cudaZGrids = tnlCuda::getNumberOfGrids( cudaBlocks.z);
   for( IndexType gridXIdx = 0; gridXIdx < cudaXGrids; gridXIdx ++ )
      for( IndexType gridYIdx = 0; gridYIdx < cudaYGrids; gridYIdx ++ )
         for( IndexType gridZIdx = 0; gridZIdx < cudaZGrids; gridZIdx ++ )
         {
            tnlTraverserGrid3DInteriorEdges< Real, Index, UserData, EntitiesProcessor, 0, 1, 0 >
                                           <<< cudaBlocks, cudaBlockSize >>>
                                          ( kernelGrid,
                                            kernelUserData,
                                            gridXIdx,
                                            gridYIdx,
                                            gridZIdx );
            checkCudaDevice;
         }
   cudaThreadSynchronize();
   
   /****
    * < 0, 0, 1 > edges
    */
   cudaBlocks.x = tnlCuda::getNumberOfBlocks( grid.getDimensions().x() + 1, cudaBlockSize.x );
   cudaBlocks.y = tnlCuda::getNumberOfBlocks( grid.getDimensions().y() + 1, cudaBlockSize.y );
   cudaBlocks.z = tnlCuda::getNumberOfBlocks( grid.getDimensions().z(), cudaBlockSize.z );
   cudaXGrids = tnlCuda::getNumberOfGrids( cudaBlocks.x );
   cudaYGrids = tnlCuda::getNumberOfGrids( cudaBlocks.y );
   cudaZGrids = tnlCuda::getNumberOfGrids( cudaBlocks.z);
   for( IndexType gridXIdx = 0; gridXIdx < cudaXGrids; gridXIdx ++ )
      for( IndexType gridYIdx = 0; gridYIdx < cudaYGrids; gridYIdx ++ )
         for( IndexType gridZIdx = 0; gridZIdx < cudaZGrids; gridZIdx ++ )
         {
            tnlTraverserGrid3DInteriorEdges< Real, Index, UserData, EntitiesProcessor, 0, 0, 1 >
                                           <<< cudaBlocks, cudaBlockSize >>>
                                          ( kernelGrid,
                                            kernelUserData,
                                            gridXIdx,
                                            gridYIdx,
                                            gridZIdx );
            checkCudaDevice;
         }
   cudaThreadSynchronize();
#endif   
}


template< typename Real,
          typename Index >
   template< typename UserData,
             typename EntitiesProcessor >
void
tnlTraverser< tnlGrid< 3, Real, tnlCuda, Index >, 0 >::
processBoundaryEntities( const GridType& grid,
                         UserData& userData ) const
{
   /****
    * Traversing boundary vertices
    */
#ifdef HAVE_CUDA
   GridType* kernelGrid = tnlCuda::passToDevice( grid );
   UserData* kernelUserData = tnlCuda::passToDevice( userData );

   dim3 cudaBlockSize( 8, 8, 4 );
   dim3 cudaBlocks;
   cudaBlocks.x = tnlCuda::getNumberOfBlocks( grid.getDimensions().x() + 1, cudaBlockSize.x );
   cudaBlocks.y = tnlCuda::getNumberOfBlocks( grid.getDimensions().y() + 1, cudaBlockSize.y );
   cudaBlocks.z = tnlCuda::getNumberOfBlocks( grid.getDimensions().z() + 1, cudaBlockSize.z );
   const IndexType cudaXGrids = tnlCuda::getNumberOfGrids( cudaBlocks.x );
   const IndexType cudaYGrids = tnlCuda::getNumberOfGrids( cudaBlocks.y );
   const IndexType cudaZGrids = tnlCuda::getNumberOfGrids( cudaBlocks.z );

   for( IndexType gridXIdx = 0; gridXIdx < cudaXGrids; gridXIdx ++ )
      for( IndexType gridYIdx = 0; gridYIdx < cudaYGrids; gridYIdx ++ )
         for( IndexType gridZIdx = 0; gridZIdx < cudaZGrids; gridZIdx ++ )
         {
            tnlTraverserGrid3DBoundaryVertices< Real, Index, UserData, EntitiesProcessor >
                                              <<< cudaBlocks, cudaBlockSize >>>
                                             ( kernelGrid,
                                               kernelUserData,
                                               gridXIdx,
                                               gridYIdx,
                                               gridZIdx );
         }
   cudaThreadSynchronize();
   checkCudaDevice;
#endif
   
}

template< typename Real,
          typename Index >
   template< typename UserData,
             typename EntitiesProcessor >
void
tnlTraverser< tnlGrid< 3, Real, tnlCuda, Index >, 0 >::
processInteriorEntities( const GridType& grid,
                         UserData& userData ) const
{
   /****
    * Traversing interior vertices
    */
#ifdef HAVE_CUDA
   GridType* kernelGrid = tnlCuda::passToDevice( grid );
   UserData* kernelUserData = tnlCuda::passToDevice( userData );

   dim3 cudaBlockSize( 8, 8, 4 );
   dim3 cudaBlocks;
   cudaBlocks.x = tnlCuda::getNumberOfBlocks( grid.getDimensions().x() + 1, cudaBlockSize.x );
   cudaBlocks.y = tnlCuda::getNumberOfBlocks( grid.getDimensions().y() + 1, cudaBlockSize.y );
   cudaBlocks.z = tnlCuda::getNumberOfBlocks( grid.getDimensions().z() + 1, cudaBlockSize.z );
   const IndexType cudaXGrids = tnlCuda::getNumberOfGrids( cudaBlocks.x );
   const IndexType cudaYGrids = tnlCuda::getNumberOfGrids( cudaBlocks.y );
   const IndexType cudaZGrids = tnlCuda::getNumberOfGrids( cudaBlocks.z );

   for( IndexType gridXIdx = 0; gridXIdx < cudaXGrids; gridXIdx ++ )
      for( IndexType gridYIdx = 0; gridYIdx < cudaYGrids; gridYIdx ++ )
         for( IndexType gridZIdx = 0; gridZIdx < cudaZGrids; gridZIdx ++ )
         {
            tnlTraverserGrid3DInteriorVertices< Real, Index, UserData, EntitiesProcessor >
                                              <<< cudaBlocks, cudaBlockSize >>>
                                             ( kernelGrid,
                                               kernelUserData,
                                               gridXIdx,
                                               gridYIdx,
                                               gridZIdx );
         }
   cudaThreadSynchronize();
   checkCudaDevice;
#endif
}


#endif /* TNLTRAVERSER_GRID3D_IMPL_H_ */
